/*

    ThermostatQA - test framework for Thermostat Monitoring Tool

    Copyright 2013 Red Hat, Inc.

This file is part of ThermostatQA

ThermostatQA is distributed under the GNU General Public License,
version 2 or any later version (with a special exception described
below, commonly known as the "Classpath Exception").

A copy of GNU General Public License (GPL) is included in this
distribution, in the file COPYING.

Linking ThermostatQA code with other modules is making a combined work
based on ThermostatQA.  Thus, the terms and conditions of the GPL
cover the whole combination.

As a special exception, the copyright holders of ThermostatQA give you
permission to link this code with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on ThermostatQA code.  If you modify ThermostatQA, you may
extend this exception to your version of the software, but you are
not obligated to do so.  If you do not wish to do so, delete this
exception statement from your version.
*/

package org.thermostat.qa.framework;

import java.awt.Rectangle;
import java.awt.image.BufferedImage;



/**
 * This class contains methods for checking if some input image contains given
 * pattern. Current algorithm can be used for sRGB images.
 * 
 * @author Pavel Tisnovsky
 */
public class ImageProcessing {

    /**
     * Threshold value for a ratio between correlation and autocorrelation values.
     * This threshold is used by the algorithm for finding some pattern in a screenshot.
     */
    private static final float CORRELATION_THRESHOLD = 0.9f;

    /**
     * Create three-dimensional array with the same size as tested image
     * dimensions (the last dimension is used for storing RGB components).
     * 
     * @param testImage
     *            tested image
     * @return newly created two-dimensional array
     */
    private static float[][][] createArrayForAllColorComponents(BufferedImage testImage)
    {
        return new float[testImage.getHeight()][testImage.getWidth()][3];
    }

    /**
     * Perform conversion from BufferedImage into three dimensional float
     * arrays. It's much faster to work with float arrays on current CPUs even
     * if it's memory inefficient. Pixels in the float array are stored with
     * offset=average_value. This means, that value of color components might
     * be negative. 
     * 
     * @param testImage
     *            tested image
     * @param array
     *            array to fill
     */
    private static void convertImageToFloatArray(BufferedImage testImage, float[][][] array)
    {
        float redAverage = 0;
        float greenAverage = 0;
        float blueAverage = 0;
        for (int y = 0; y < testImage.getHeight(); y++)
        {
            for (int x = 0; x < testImage.getWidth(); x++)
            {
                // get the RGBA values packed in one integer
                int c = testImage.getRGB(x, y);
                // filter out alpha channel
                c = c & 0xffffff;
                redAverage += ((c >> 16) & 0xff);
                greenAverage += ((c >> 8) & 0xff);
                blueAverage += (c & 0xff);
            }
        }
        int pixels = testImage.getWidth() * testImage.getHeight();
        redAverage /= pixels;
        greenAverage /= pixels;
        blueAverage /= pixels;
        //System.out.println("red: " + redAverage);
        //System.out.println("green: " + greenAverage);
        //System.out.println("blue: " + blueAverage);

        for (int y = 0; y < testImage.getHeight(); y++)
        {
            for (int x = 0; x < testImage.getWidth(); x++)
            {
                // get the RGBA values packed in one integer
                int c = testImage.getRGB(x, y);
                // filter out alpha channel
                c = c & 0xffffff;
                array[y][x][0] = ((c >> 16) & 0xff) - redAverage;
                array[y][x][1] = ((c >> 8) & 0xff) - greenAverage;
                array[y][x][2] = (c & 0xff) - blueAverage;
            }
        }
    }

    /**
     * Tries to find given pattern in an input image. Only the exact pattern
     * could be found by this method.
     * 
     * @param pattern
     *            given pattern stored represented as BufferedImage
     * @param inputImage
     *            input image
     * @return rectangle containing area where the pattern has been found, or
     *         null
     */
    public static Rectangle findPattern(BufferedImage pattern, BufferedImage inputImage){
        Rectangle actionArea = new Rectangle(0, 0, inputImage.getWidth(), inputImage.getHeight());
        Rectangle result = new Rectangle(0, 0, 0, 0);
        boolean found = false;
        boolean ok = true;

        PATTERN_FOUND:
        for (int x = actionArea.x; (x < (actionArea.x + actionArea.width - pattern.getWidth())) && !found; x++)
        {
            for (int y = actionArea.y; (y < (actionArea.y + actionArea.height - pattern.getHeight())) && !found; y++)
            {
                for (int mx = 0; (mx < pattern.getWidth()) && ok; mx++)
                {
                    for (int my = 0; (my < pattern.getHeight()) && ok; my++)
                    {
                        if (pattern.getRGB(mx, my) != inputImage.getRGB(x + mx, y + my))
                        {
                            ok = false;
                        }
                    }
                }
                if (ok)
                {
                    found = true;
                    result.x = x;
                    result.y = y;
                    result.height = pattern.getHeight();
                    result.width = pattern.getWidth();
                    break PATTERN_FOUND;
                }
                ok = true;
            }
        }

        // if the pattern has not been found, return null instead of the proper rectangle
        return found ? result : null;
    }

    /**
     * Tries to find given pattern in an input image.
     * 
     * @param pattern
     *            given pattern stored represented as BufferedImage
     * @param inputImage
     *            input image
     * @return rectangle containing area where the pattern have been found, or
     *         null
     */
    protected static Rectangle findBlurredPattern(BufferedImage pattern, BufferedImage inputImage)
    {
        int maxX = inputImage.getWidth() - pattern.getWidth() - 1;
        int maxY = inputImage.getHeight() - pattern.getHeight() - 1;
        int markerMaxX = pattern.getWidth();
        int markerMaxY = pattern.getHeight();

        // it is much faster to work directly with color components stored as float values
        float[][][] testImageArray = createArrayForAllColorComponents(inputImage);
        float[][][] markerImageArray = createArrayForAllColorComponents(pattern);

        //System.out.println("Starting conversion");
        convertImageToFloatArray(inputImage, testImageArray);
        convertImageToFloatArray(pattern, markerImageArray);
        //System.out.println("Conversion done");

        // autocorellation value is needed to compute and compare correlation values
        // and to find the exact pattern (if it exists in the input image)
        double autocorrelation = computeCorrelation(markerMaxX, markerMaxY, markerImageArray, markerImageArray, 0, 0);

        int bestX = -1;
        int bestY = -1;
        double bestCorrelation = -1;
        boolean foundExact = false;

        for (int yoffset = 0; (yoffset < maxY) && !foundExact; yoffset++ )
        {
            //System.out.println("Processing line: " + yoffset + " of " + (maxY - 1));
            for (int xoffset = 0; (xoffset < maxX) && !foundExact; xoffset++)
            {
                double correlation = computeCorrelation(markerMaxX, markerMaxY, testImageArray, markerImageArray, xoffset, yoffset);
                if (correlation > bestCorrelation)
                {
                    bestCorrelation = correlation;
                    bestX = xoffset;
                    bestY = yoffset;
                    if (correlation == autocorrelation)
                    {
                        foundExact = true;
                    }
                }
            }
        }

        //System.out.println(autocorrelation);
        //System.out.println(bestCorrelation);
        //System.out.println(bestX + " " + bestY);
        if (bestCorrelation / autocorrelation < CORRELATION_THRESHOLD)
        {
            return null;
        }
        // pattern has been found!
        return new Rectangle(bestX, bestY, pattern.getWidth(), pattern.getHeight());
    }

    /**
     * Compute correlation for given two images and a 2D offset.
     * 
     * @param maxX
     * @param maxY
     * @param testImageArray
     *            first input array
     * @param markerImageArray
     *            second input array
     * @param xoffset
     *            horizontal part of an offset
     * @param yoffset
     *            vertical part of an offset
     * @return correlation of two arrays with the given offset
     */
    private static double computeCorrelation(int maxX, int maxY, float[][][] testImageArray,
                    float[][][] markerImageArray, int xoffset, int yoffset)
    {
        double correlation = 0;
        for (int y = 0; y < maxY; y++)
        {
            for (int x = 0; x < maxX; x++)
            {
                for (int rgbIndex = 0; rgbIndex < 3; rgbIndex++)
                {
                    float colorComponent1 = markerImageArray[y][x][rgbIndex];
                    float colorComponent2 = testImageArray[yoffset + y][xoffset + x][rgbIndex];
                    correlation += colorComponent1 * colorComponent2;
                }
            }
        }
        return correlation;
    }

}
